/**
CUDA Audio processing application.
CS 441/641 Take-Home Final Exam Code, problem 2.

If you do have CUDA installed:
  nvcc 2_cuda_sound.cu

If you don't have CUDA, paste this into NetRun with:
  Language: CUDA
  Mode: Whole Program (main)

Dr. Orion Lawlor, lawlor@alaska.edu, 2014-12-10 (Public Domain)
*/
#include <iostream>
#include <iomanip>
#include <cuda.h>
#include "lib/inc.c" /* netrun timing stuff, time_in_seconds() */

/**
 A soundSample represents the sound at one moment in time.
*/
class soundSample {
public:
	float L,R; // left and right audio channels: sound pressure levels

	// Default constructor: initialize to silence
	__device__ __host__ soundSample() :L(0.0), R(0.0) {}

	// Create a soundSample from two values:
	__device__ __host__ soundSample(float L_,float R_) :L(L_), R(R_) {}


	// Add this soundSample to us:
	__device__ __host__ void add(const soundSample &s) {
		L+=s.L; R+=s.R;
	}

	// Atomically add this soundSample to us.  Works across multiple threads.
	__device__ void atomicAdd(const soundSample &s) {
		::atomicAdd(&L,s.L);
		::atomicAdd(&R,s.R);
	}

	// Return true if these two sounds are similar
	bool similarTo(const soundSample &s) const {
		return (abs(L-s.L)+abs(R-s.R))<0.0001;
	}
};

/* Compute the sound contribution at this time, of this frequency. */
__device__ __host__ //<- this function runs on CPU or GPU
soundSample getSound(int time,int freq,float amp,float phaseShift) {
	double freqHz=40+20.0*freq; // sound frequency, in cycles/second
	double freqRads=freqHz*2.0*M_PI; // sound frequency, in radians/sec
	double audioRate=44.1e3; // 4.1 KHz audio (CD quality), samples/sec
	double timePerSample=1.0/audioRate; // seconds/sample
	double timeSec=time*timePerSample; // time in seconds 
	float phase=timeSec*freqRads; // current sound phase, in radians
	return soundSample(sin(phase)*amp,sin(phase+phaseShift)*amp);
}

/* GPU kernel: fill this array with sound.  
   Each thread adds a different frequency to the output.
*/
__global__ void make_sound(soundSample *sound,
	int ntime,int nfreq,
	float amp,float phaseShift) 
{
	int thread=threadIdx.x + blockIdx.x*blockDim.x; // <- my thread index

	// Zero the sound array first
	if (thread==0)
		for (int time=0;time<ntime;time++) {
			sound[time]=soundSample();
		}
	__syncthreads(); // wait until the whole array is zeroed out

	// Add my octave's contribution to the whole sound
	for (int time=0;time<ntime;time++) {
		sound[time].atomicAdd(getSound(time,thread,amp,phaseShift));
	}
}

// The "chk" function checks for runtime errors, and shows performance.
void chk(const char *whereInCode) {
	static double start=time_in_seconds();
	if (cudaGetLastError()!=cudaSuccess) 
		std::cout<<"CUDA failure at "<<whereInCode<<std::endl;
	double cur=time_in_seconds();
	std::cout<<"to "<<whereInCode<<": "
		<<std::fixed<<std::setprecision(3)
		<<(cur-start)*1.0e3<<" ms"<<std::endl;
	start=cur;
}

int main(int argc,char *argv[]) 
{
	chk("Startup");
// Allocate space shared between CPU and GPU
	int ntime=16*1024; // total number of sound samples
	int nfreq=32; // frequencies audible at each sample
	float amp=0.01; // amplitude of each frequency
	float phaseShift=0.1234; // per-ear phase shift, radians
	
	soundSample *sound; // shared array of n values 
	cudaMallocManaged( &sound, ntime*sizeof(soundSample) ); 
	chk("cudaMalloc");

// Run GPU kernel to make sound
	make_sound<<<1,nfreq>>>(sound,ntime,nfreq,
		0.01,0.1234); /* run kernel on n GPU threads */ 

	cudaDeviceSynchronize(); /* Wait for kernel to finish filling vals array */ 
	chk("kernel");

// Spot-check results
	for (int time=0;time<ntime;time+=800) {
		soundSample s; // zero initialized
		for (int freq=0;freq<nfreq;freq++)
			s.add(getSound(time,freq,amp,phaseShift));
		soundSample sGPU=sound[time];
		std::cout<<"time "<<time<<": ";
		if (s.similarTo(sGPU)) std::cout<<"OK\n";
		else std::cout<<" MISMATCH!\n";
	}
	chk("readback");
        return 0;
}

